/*
 * esteid.c: EstEID functions
 *
 * Copyright (C) 2001, 2002  Juha Yrjl <juha.yrjola@iki.fi>
 * Copyright (C) 2002  Priit Randla <priit.randla@eyp.ee>
 * Copyright (C) 2002  Marie Fischer <marie@vtl.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "internal.h"
#include "esteid.h"
#include "asn1.h"
#include "log.h"
#include "card-mcrd.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <sys/stat.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

struct sc_esteid_card * sc_esteid_card_new()
{
	struct sc_esteid_card *estcard;

	estcard = (struct sc_esteid_card *) calloc(1,sizeof(struct sc_esteid_card));
	if (estcard == NULL)
		return NULL;
	memset(estcard, 0, sizeof(struct sc_esteid_card));
	estcard->magic = SC_ESTEID_CARD_MAGIC;
	return estcard;
}

void sc_esteid_card_free(struct sc_esteid_card *estcard)
{
	int recno=0;
	
	assert(estcard != NULL && estcard->magic == SC_ESTEID_CARD_MAGIC);

	estcard->magic = 0;
	if (estcard->label)
		free(estcard->label);
	if (estcard->serial_number)
		free(estcard->serial_number);
	if (estcard->manufacturer_id)
		free(estcard->manufacturer_id);
	
	for ( recno = 1; recno <= SC_ESTEID_PUBLIC_LINE_COUNT; recno++) {
		if ( estcard->public_info[recno-1] )
			free(estcard->public_info[recno-1]); 
	}
	for (recno=0;recno < estcard->pin_count;recno++){
		if ( estcard->pins[recno] )
			free(estcard->pins[recno]); 
	}

	for (recno=0;recno < estcard->cert_count;recno++){
		if ( estcard->certs[recno] )
			free(estcard->certs[recno]); 
	}

	for (recno=0;recno < estcard->prkey_count;recno++){
		if ( estcard->prkeys[recno] )
			free(estcard->prkeys[recno]); 
	}

	free(estcard);
}

int sc_esteid_bind(struct sc_card *card,
		   struct sc_esteid_card **estcard_out)
{
	int err, recno=0;
	struct sc_esteid_card *estcard = NULL;
	struct sc_path tmppath;
	struct sc_context *ctx;
	char databuf[SC_ESTEID_PUBLIC_MAXLINELEN + 1];

	assert(sc_card_valid(card) && estcard_out != NULL);
	ctx = card->ctx;
	SC_FUNC_CALLED(ctx, 1);
	estcard = sc_esteid_card_new();
	if (estcard == NULL)
		return SC_ERROR_OUT_OF_MEMORY;
	estcard->card = card;

	/* FIXME: parse config file */
	estcard->opts.use_cache = 1;

	err = sc_lock(card);
	if (err) {
		error(ctx, "sc_lock() failed: %s\n", sc_strerror(err));
		goto error;
	}
	

	sc_format_path("3F00", &tmppath);
	tmppath.type = SC_PATH_TYPE_FILE_ID;

	err = sc_select_file(card, &tmppath, NULL);
	if (err) {
		error(ctx, "Error selecting EstEID MF(DIR): %s\n", sc_strerror(err));
		err = SC_ERROR_ESTEID_NOT_FOUND;
		goto error;
	}

	sc_format_path("EEEE", &tmppath);
	tmppath.type = SC_PATH_TYPE_FILE_ID;
		
	err = sc_select_file(card, &tmppath, NULL);
	if (err) {
		error(ctx, "Error selecting EstEID DF(DIR): %s\n", sc_strerror(err));
		err = SC_ERROR_ESTEID_NOT_FOUND;
		goto error;
	}

	sc_format_path("5044", &tmppath);
	tmppath.type = SC_PATH_TYPE_FILE_ID;;
	err = sc_select_file(card, &tmppath, NULL);
	if (err) {
		error(ctx, "Error selecting public info file 'PD'): %s\n", sc_strerror(err));
		err = SC_ERROR_ESTEID_NOT_FOUND;
		goto error;
	}
	
	for ( recno = 1 ; recno <= SC_ESTEID_PUBLIC_LINE_COUNT; recno++) {
		memset( databuf, 0, SC_ESTEID_PUBLIC_MAXLINELEN + 1);
		err = sc_read_record( card, recno, databuf, SC_ESTEID_PUBLIC_MAXLINELEN, 0);
		if (err < 0) {
			error(ctx,"READ RECORD failed: %s\n", sc_strerror(err));
			err = SC_ERROR_ESTEID_NOT_FOUND;
			goto error;
		}
		estcard->public_info[recno-1] = (char*) malloc( strlen(databuf)+1);

		if ( estcard->public_info[recno-1] == NULL ) {
			error(ctx,"Couldn't allocate memory?!");
			err = SC_ERROR_ESTEID_NOT_FOUND;
			goto error;
		}

		memcpy( estcard->public_info[recno-1],  databuf, strlen(databuf)+1);
	}

	//from eyp
	//no tokeninfo on esteid...
	estcard->manufacturer_id = strdup("Orga");
	estcard->serial_number = strdup(estcard->public_info[DOCUMENT_NR]);
	estcard->label = strdup("EstEID");

	
	for(recno=0;recno<SC_ESTEID_MAX_PINS;recno++){
		estcard->pins[recno] = (struct sc_esteid_object *) malloc(sizeof(struct sc_esteid_object));
	}	

	for(recno=0;recno<SC_ESTEID_MAX_PRKEYS;recno++){
		estcard->prkeys[recno] = (struct sc_esteid_object *) malloc(sizeof(struct sc_esteid_object));
	}

	for(recno=0;recno<SC_ESTEID_MAX_CERTS;recno++){
		estcard->certs[recno] = (struct sc_esteid_object *) malloc(sizeof(struct sc_esteid_object));
	}

	*estcard_out = estcard;
	sc_unlock(card);
	return 0;
error:
	sc_esteid_card_free(estcard);
	sc_unlock(card);
	SC_FUNC_RETURN(ctx, 1, err);

}

int sc_esteid_unbind(struct sc_esteid_card *estcard)
{
	assert(estcard != NULL && estcard->magic == SC_ESTEID_CARD_MAGIC);
	SC_FUNC_CALLED(estcard->card->ctx, 1);
	sc_esteid_card_free(estcard);
	return 0;
}

// convert difficult chars (in card->certs->label) to ascii
// to work around bug in mozilla...
static u8 * convert_string(u8 *s)
{
        int i;
        u8 *t;
        t=strdup(s);
        for(i=0;i<strlen(t);i++){
		if(t[i]<0x20 || t[i]>0x7c){
			switch (t[i]) {
			case 0xc4:
				t[i]='A';
				break;
			case 0xd6:
				t[i]='O';
				break;
			case 0xdc:
				t[i]='U';
				break;
			case 0xd5:
				t[i]='O';
				break;
			default:
				t[i]='?';
				break;
			}
		}
        }
        return t;
}


int sc_esteid_enum_certificates(struct sc_esteid_card *card, int application)
{
	struct sc_esteid_cert_info *cert_info[2];
	int i;
	int certcount=2;	//1 - pkcs11, 2 - other

	
	assert(card != NULL);
	debug(card->card->ctx, "called\n");
	
	if (application == SC_ESTEID_APP_PKCS11) {
		certcount = 1;
	} else if(application==SC_ESTEID_APP_OTHER) {
		certcount = 2;
	}

	if (card->cert_count){
		debug(card->card->ctx, "certs already enumerated\n");
		return card->cert_count;	/* already enumerated */
	}
	card->cert_count = certcount;

	for(i=0;i<certcount;i++){
		cert_info[i] = (struct sc_esteid_cert_info*) calloc(1, sizeof(struct sc_esteid_cert_info));
	}


	sc_format_path( "AACE", &cert_info[0]->path);
	cert_info[0]->path.type = SC_PATH_TYPE_FILE_ID;
	snprintf( cert_info[0]->label, SC_ESTEID_MAX_LABEL_SIZE, "%s, %s",card->public_info[SURNAME], card->public_info[GIVEN_NAMES1]);
	cert_info[0]->id.value[0] = 0x42;
	cert_info[0]->id.len = 1;
	card->certs[0]->type = SC_ESTEID_TYPE_CERT;
//	snprintf(card->certs[0]->label, SC_ESTEID_MAX_LABEL_SIZE, "firstcert");
	snprintf(card->certs[0]->label, SC_ESTEID_MAX_LABEL_SIZE, "%s,%s", convert_string(card->public_info[SURNAME]), convert_string(card->public_info[GIVEN_NAMES1]));

	card->certs[0]->flags = 0;
	card->certs[0]->auth_id.value[0] = 0x42;
	card->certs[0]->auth_id.len = 1;
	card->certs[0]->data = cert_info[0];

	if (certcount > 1) {
		sc_format_path( "DDCE", &cert_info[1]->path);
		cert_info[1]->path.type = SC_PATH_TYPE_FILE_ID;	
		snprintf( cert_info[1]->label, SC_ESTEID_MAX_LABEL_SIZE, "%s, %s",card->public_info[SURNAME], card->public_info[GIVEN_NAMES1]);
		cert_info[1]->id.value[0] = 0x43;
		cert_info[1]->id.len = 1;
		card->certs[1]->type = SC_ESTEID_TYPE_CERT;
//		snprintf(card->certs[1]->label, SC_ESTEID_MAX_LABEL_SIZE, "secondcert");
		snprintf(card->certs[1]->label, SC_ESTEID_MAX_LABEL_SIZE, "%s,%s", convert_string(card->public_info[SURNAME]), convert_string(card->public_info[GIVEN_NAMES1]));

		card->certs[1]->flags = 0;
		card->certs[1]->auth_id.value[0] = 0x43;
		card->certs[1]->auth_id.len = 1;
		card->certs[1]->data = cert_info[1];
	}
	
	return card->cert_count;
}

int sc_esteid_find_cert_by_id(struct sc_esteid_card *card,
				const struct sc_esteid_id *id,
				struct sc_esteid_object **out)
{
	int r, i;
	debug(card->card->ctx, "called\n");
	
	r = sc_esteid_enum_certificates(card, SC_ESTEID_APP_OTHER);
	if (r < 0)
		return r;
	for (i = 0; i < card->cert_count; i++) {
		struct sc_esteid_cert_info *cert = card->certs[i]->data;
		if (sc_esteid_compare_id(&cert->id, id) == 1) {
			*out = card->certs[i];
			return 0;
		}
	}
	return SC_ERROR_OBJECT_NOT_FOUND;
}

int sc_esteid_enum_private_keys(struct sc_esteid_card *card, int application)
{
	struct sc_esteid_prkey_info *prkey_info[2];
	int i;
	int prkeycount=2;

	debug(card->card->ctx, "called\n");
	assert(card != NULL);

	if (card->prkey_count) {
		debug(card->card->ctx, "prkeys already enumerated\n");
		return card->prkey_count;
	}

	if (application == SC_ESTEID_APP_PKCS11) {
		prkeycount = 1;
	} else if(application==SC_ESTEID_APP_OTHER) {
		prkeycount = 2;
	}


	
	card->prkey_count = prkeycount; 

	for(i=0;i<prkeycount;i++){
		prkey_info[i] = (struct sc_esteid_prkey_info*) calloc(1, sizeof(struct sc_esteid_prkey_info));
	}

	snprintf( prkey_info[0]->label, SC_ESTEID_MAX_LABEL_SIZE, "%s, %s", card->public_info[SURNAME], card->public_info[GIVEN_NAMES1]);
	prkey_info[0]->id.value[0] = 0x42;
	prkey_info[0]->id.len = 1;    
	prkey_info[0]->key_reference = 1;    /* PIN1 */
	sc_format_path( "1100", &prkey_info[0]->path);
	prkey_info[0]->modulus_length = 1024;
	prkey_info[0]->usage =  SC_ESTEID_PRKEY_USAGE_DECRYPT| 
					  SC_ESTEID_PRKEY_USAGE_SIGN |
					  SC_ESTEID_PRKEY_USAGE_UNWRAP; 
	card->prkeys[0]->type = SC_ESTEID_TYPE_PRKEY;
//	snprintf(card->prkeys[0]->label, SC_ESTEID_MAX_LABEL_SIZE, "first prkey");	//just so there is something
	snprintf(card->prkeys[0]->label, SC_ESTEID_MAX_LABEL_SIZE, "%s,%s", convert_string(card->public_info[SURNAME]), convert_string(card->public_info[GIVEN_NAMES1]));

	card->prkeys[0]->flags = 0;
	card->prkeys[0]->auth_id.value[0] = 0x42;
	card->prkeys[0]->auth_id.len = 1;
	card->prkeys[0]->data = prkey_info[0];
	
	if (prkeycount > 1) {
		snprintf( prkey_info[1]->label, SC_ESTEID_MAX_LABEL_SIZE, "%s, %s", card->public_info[SURNAME], card->public_info[GIVEN_NAMES1]);
		prkey_info[1]->id.value[0] = 0x43;
		prkey_info[1]->id.len = 1;    
		prkey_info[1]->key_reference = 2;    /* PIN2 */
		sc_format_path( "0100", &prkey_info[1]->path);
		prkey_info[1]->modulus_length = 1024;
		prkey_info[1]->usage = SC_ESTEID_PRKEY_USAGE_NONREPUDIATION;
		card->prkeys[1]->type = SC_ESTEID_TYPE_PRKEY;
//		snprintf(card->prkeys[1]->label, SC_ESTEID_MAX_LABEL_SIZE, "second prkey");	//just so there is something
		snprintf(card->prkeys[1]->label, SC_ESTEID_MAX_LABEL_SIZE, "%s,%s", convert_string(card->public_info[SURNAME]), convert_string(card->public_info[GIVEN_NAMES1]));
		card->prkeys[1]->flags = 0;
		card->prkeys[1]->auth_id.value[0] = 0x43;
		card->prkeys[1]->auth_id.len = 1;
		card->prkeys[1]->data = prkey_info[1];
	}

	return card->prkey_count;

}

int sc_esteid_find_prkey_by_id(struct sc_esteid_card *card,
				const struct sc_esteid_id *id,
//				struct sc_esteid_prkey_info **key_out){
				struct sc_esteid_object **out){
	int r, i;
	debug(card->card->ctx, "called\n");	
	r = sc_esteid_enum_private_keys(card, SC_ESTEID_APP_OTHER);
	if (r < 0)
		return r;
	for (i = 0; i < card->prkey_count; i++) {
		struct sc_esteid_prkey_info *key = card->prkeys[i]->data;
		if (sc_esteid_compare_id(&key->id, id) == 1) {
//			*key_out = key;
			*out = card->prkeys[i];
			return 0;
		}
	}
	return SC_ERROR_OBJECT_NOT_FOUND;

}

int sc_esteid_enum_pins(struct sc_esteid_card *estcard, int application)
{
	int r;
	struct sc_context *ctx = estcard->card->ctx;
	u8 databuf[256];	//fimxe -- mf
//	u8 databuf[SC_MAX_APDU_BUFFER_SIZE];
	struct sc_path path;
	int magic_number = 5;
	int i;
	int pincount=2;

	struct sc_esteid_pin_info *pin_info[3];

	assert(estcard != NULL);
	SC_FUNC_CALLED(ctx, 1);


	if (application == SC_ESTEID_APP_PKCS11) {
		pincount = 1;
	} else if(application==SC_ESTEID_APP_OTHER) {
		pincount = 3;
	}


	for(i=0;i<pincount;i++){
		pin_info[i] = (struct sc_esteid_pin_info*) calloc(1, sizeof(struct sc_esteid_pin_info));
	}

	//fixme - do we need this?
	r = sc_lock(estcard->card);
	SC_TEST_RET(estcard->card->ctx, r, "sc_lock() failed");

	sc_format_path("3F00", &path);
	path.type = SC_PATH_TYPE_FILE_ID;
	r = sc_select_file( estcard->card, &path, NULL);
	if (r) {
		error(ctx, "Error selecting EstEID MF(DIR): %s\n", sc_strerror(r));
		return 0;
	}
	sc_format_path("0016", &path);
	path.type = SC_PATH_TYPE_FILE_ID;
	r = sc_select_file( estcard->card, &path, NULL);
	if (r) {
		error(ctx, "Error selecting file  '0016': %s\n", sc_strerror(r));
		return 0;
	}

	estcard->pin_count = pincount;		//fixme
	for (i=0;i<pincount;i++){
		pin_info[i]->magic = SC_ESTEID_PIN_MAGIC;
		if (i==0){
			pin_info[i]->auth_id.value[0] = 0x42;
			pin_info[i]->min_length = 4;
			pin_info[i]->reference=1;
		}
		if (i==1) {
			pin_info[i]->auth_id.value[0] = 0x43;
			pin_info[i]->min_length = 5;
			pin_info[i]->reference=2;
		}
		if (i==2) {
			pin_info[i]->auth_id.value[0] = 0x44;
			pin_info[i]->min_length = 8;
			pin_info[i]->reference=0;
		}
		pin_info[i]->auth_id.len = 1;
		pin_info[i]->stored_length = 12;
		pin_info[i]->pad_char = ' ';
		pin_info[i]->verified = 0;


		r = sc_read_record( estcard->card, i+1, databuf, sizeof(databuf),  0);
		if (r < 0) {
			error(ctx, "READ RECORD from '0016' failed: %s\n", sc_strerror(r));
			return 0;
		}
		pin_info[i]->tries_left = databuf[magic_number];
		
		estcard->pins[i]->type = SC_ESTEID_TYPE_AUTH;
		estcard->pins[i]->flags = 0;
		estcard->pins[i]->auth_id.value[0] = i+1;
		estcard->pins[i]->data=pin_info[i];

	}	
	sprintf(estcard->pins[0]->label,"autentimine - PIN1");
	if(pincount>1)
		sprintf(estcard->pins[1]->label,"allkirjastamine - PIN2");
	if(pincount>2)
		sprintf(estcard->pins[2]->label,"PUK");
		estcard->pins[2]->flags |= SC_ESTEID_PIN_FLAG_UNBLOCKING_PIN;

//	r=sc_unlock(estcard->card);
	sc_unlock(estcard->card);

	//fixme - what do we check here?
	if (r <= 0)
		return r;
	return estcard->pin_count;
}


int sc_esteid_find_pin_by_auth_id(struct sc_esteid_card *card,
					const struct sc_esteid_id *id,
					struct sc_esteid_object **out){
	int r, i;
	debug(card->card->ctx, "called\n");	
	
	r = sc_esteid_enum_pins(card, SC_ESTEID_APP_OTHER);
	if (r < 0)
		return r;
	for (i = 0; i < card->pin_count; i++) {
		struct sc_esteid_pin_info *pin = card->pins[i]->data;
		if (sc_esteid_compare_id(&pin->auth_id, id) == 1) {
			*out = card->pins[i];
			return 0;
		}
	}
	return SC_ERROR_OBJECT_NOT_FOUND;
}

int sc_esteid_read_file(struct sc_esteid_card *estcard,
			const struct sc_path *path,
			u8 **buf, size_t *buflen,
			struct sc_file **file_out)
{
	struct sc_file *file;
	u8	*data = NULL;
	size_t	len = 0;
	int	r = -1;

	assert(estcard != NULL && path != NULL && buf != NULL);
	SC_FUNC_CALLED(estcard->card->ctx, 1);
	if (estcard->opts.use_cache) {
		r = sc_esteid_read_cached_file(estcard, path, &data, &len);
	}
	if (r) {
		r = sc_lock(estcard->card);
		SC_TEST_RET(estcard->card->ctx, r, "sc_lock() failed");
		r = sc_select_file(estcard->card, path, &file);
		if (r) {
			sc_unlock(estcard->card);
			return r;
		}
		len = file->size;
		if (file_out != NULL)
			*file_out = file;
		else
			sc_file_free(file);
		data = (u8 *) malloc(len);
		if (data == NULL) {
			sc_unlock(estcard->card);
			return SC_ERROR_OUT_OF_MEMORY;
		}
		r = sc_read_binary(estcard->card, 0, data, len, 0);
		if (r < 0) {
			sc_unlock(estcard->card);
			free(data);
			return r;
		}
		sc_unlock(estcard->card);
	}
	*buf = data;
	*buflen = len;
	return 0;
}

int sc_esteid_compare_id(const struct sc_esteid_id *id1,
			 const struct sc_esteid_id *id2)
{
	assert(id1 != NULL && id2 != NULL);
	if (id1->len != id2->len)
		return 0;
	return memcmp(id1->value, id2->value, id1->len) == 0;
}

void sc_esteid_format_id(const char *str, struct sc_esteid_id *id)
{
	int len = 0;
	u8 *p = id->value;

	while (*str) {
		int byte;
		
		if (sscanf(str, "%02X", &byte) != 1)
			break;
		*p++ = byte;
		len++;
		str += 2;
	}
	id->len = len;
	return;
}

void sc_esteid_print_id(const struct sc_esteid_id *id)
{
	int i;

	for (i = 0; i < id->len; i++)
		printf("%02X", id->value[i]);
}

int sc_esteid_hex_string_to_id(const char *in, struct sc_esteid_id *out)
{
	out->len = sizeof(out->value);
	return sc_hex_to_bin(in, out->value, &out->len);
}


//untested -- mf
static int esteid_prepare_decipher( struct sc_esteid_card *estcard, 
				    const struct sc_esteid_prkey_info *prkey)
{
	struct sc_context *ctx = estcard->card->ctx;
	int r;

	r = sc_restore_security_env(estcard->card, 6); /* empty SE */
	SC_TEST_RET(ctx, r, "sc_restore_security_env() failed");
	r = mcrd_delete_ref_to_authkey(estcard->card);
	SC_TEST_RET(ctx, r, "mcrd_delete_ref_to_authkey() failed");
	r = mcrd_delete_ref_to_signkey(estcard->card);
	SC_TEST_RET(ctx, r, "mcrd_delete_ref_to_signkey() failed");
	r = mcrd_set_decipher_key_ref(estcard->card, prkey->key_reference);
	SC_TEST_RET(ctx, r, "mcrd_set_decipher_key_ref() failed");
 	
	return 0;
}

//untested - mf
int sc_esteid_decipher(struct sc_esteid_card *estcard,
		       const struct sc_esteid_prkey_info *prkey,
		       const u8 * in, int inlen, u8 *out, int outlen)
{
	int r;
	struct sc_security_env senv;
	struct sc_context *ctx = estcard->card->ctx;
	
	senv.algorithm_ref = 0x02;
	senv.file_ref = prkey->path;
	senv.operation = SC_SEC_OPERATION_DECIPHER;
	senv.key_ref[0] = prkey->key_reference;
	r = esteid_prepare_decipher( estcard, prkey);	
	r = sc_decipher(estcard->card, in, inlen, out, outlen);
	SC_TEST_RET(ctx, r, "sc_decipher() failed");
	return r;
}

int esteid_prepare_sign( struct sc_esteid_card *estcard, 
				const struct sc_esteid_prkey_info *prkey, int hash)
{
	struct sc_context *ctx = estcard->card->ctx;
	int r;

	r = sc_restore_security_env(estcard->card, 1); /* empty SE */
	SC_TEST_RET(ctx, r, "sc_restore_security_env() failed");

	return 0; 
}
	

int sc_esteid_compute_signature(struct sc_esteid_card *estcard,
				const struct sc_esteid_prkey_info *prkey,
				int hash, const u8 *in, int inlen, u8 *out,
				int outlen)
{
	int r;
	struct sc_context *ctx = estcard->card->ctx;
	struct sc_path tmppath;

	debug(estcard->card->ctx, "called\n");	

	//select df ... just in case - mf
	
	sc_format_path("3F00", &tmppath);
	tmppath.type = SC_PATH_TYPE_FILE_ID;

	r = sc_select_file(estcard->card, &tmppath, NULL);
	if (r) {
		error(estcard->card->ctx, "Error selecting EstEID MF(DIR): %s\n", sc_strerror(r));
		r = SC_ERROR_ESTEID_NOT_FOUND;
		SC_TEST_RET(ctx, r, "sc_compute_signature() failed");
		return r;
	}

	sc_format_path("EEEE", &tmppath);
	tmppath.type = SC_PATH_TYPE_FILE_ID;
		
	r = sc_select_file(estcard->card, &tmppath, NULL);
	if (r) {
		error(estcard->card->ctx, "Error selecting EstEID DF(DIR): %s\n", sc_strerror(r));
		r = SC_ERROR_ESTEID_NOT_FOUND;
		SC_TEST_RET(ctx, r, "sc_compute_signature() failed");
		return r;
	}	
	
	


	r = esteid_prepare_sign( estcard, prkey, hash); 
	SC_TEST_RET(ctx, r, "esteid_prepare_sign() failed");


	//fixme: this is ugly but seems to work -- mf
	switch ( prkey->key_reference ) {
	case 1:
		r = mcrd_encrypt( estcard->card, in, inlen, out, outlen);

		break;
	case 2:
		r = sc_compute_signature( estcard->card, in, inlen, out, outlen);
		break;
	}
	SC_TEST_RET(ctx, r, "sc_compute_signature() failed");
	return r;

}

int select_esteid_df( struct sc_esteid_card *card)
{
	struct sc_path path;
	int r;

	sc_format_path("3F00", &path);
	path.type = SC_PATH_TYPE_FILE_ID;
	r = sc_select_file( card->card, &path, NULL);
	if (r) {
		debug(card->card->ctx, "Select MF failed?...\n");
		return r;
	}

	sc_format_path("EEEE", &path);
	path.type = SC_PATH_TYPE_FILE_ID;
	r = sc_select_file( card->card, &path, NULL);
	if ( r ) {
		debug(card->card->ctx, "Select EEEE succeeded...\n");
		return r;
	}
	return 0;
};



static int generate_cache_filename(struct sc_esteid_card *estcard,
				   const struct sc_path *path,
				   char *buf, size_t bufsize)
{
	char dir[80];
        char pathname[SC_MAX_PATH_SIZE*2+1];
	int i, r;
        const u8 *pathptr;
        size_t pathlen;

	if (path->type != SC_PATH_TYPE_PATH)
                return SC_ERROR_INVALID_ARGUMENTS;
	assert(path->len <= SC_MAX_PATH_SIZE);
	r = sc_get_cache_dir(estcard->card->ctx, dir, sizeof(dir));
	if (r)
		return r;
	pathptr = path->value;
	pathlen = path->len;
	if (pathlen > 2 && memcmp(pathptr, "\x3F\x00", 2) == 0) {
                pathptr += 2;
		pathlen -= 2;
	}
	for (i = 0; i < pathlen; i++)
                sprintf(pathname + 2*i, "%02X", pathptr[i]);
	r = snprintf(buf, bufsize, "%s/%s_%s_%s_%s", dir,
		     estcard->manufacturer_id, estcard->label,
		     estcard->serial_number, pathname);
	if (r < 0)
		return SC_ERROR_BUFFER_TOO_SMALL;
        return 0;
}

int sc_esteid_read_cached_file(struct sc_esteid_card *estcard,
			       const struct sc_path *path,
			       u8 **buf, size_t *bufsize)
{
	char fname[160];
	int r;
	FILE *f;
	size_t c;
	struct stat stbuf;
	u8 *data = NULL;

	r = generate_cache_filename(estcard, path, fname, sizeof(fname));
	if (r != 0)
		return r;
	r = stat(fname, &stbuf);
	if (r)
		return SC_ERROR_FILE_NOT_FOUND;
	c = stbuf.st_size;
	if (*buf == NULL) {
		data = (u8 *) malloc(stbuf.st_size);
		if (data == NULL)
			return SC_ERROR_OUT_OF_MEMORY;
	} else
		if (c > *bufsize)
			return SC_ERROR_BUFFER_TOO_SMALL;
	f = fopen(fname, "r");
	if (f == NULL) {
		if (data)
			free(data);
		return SC_ERROR_FILE_NOT_FOUND;
	}
	if (data)
		*buf = data;
	c = fread(*buf, 1, c, f);
        fclose(f);
	if (c != stbuf.st_size) {
		if (data)
			free(data);
		return SC_ERROR_BUFFER_TOO_SMALL;
	}
	*bufsize = c;
	if (data)
		*buf = data;
        return 0;
}

int sc_esteid_cache_file(struct sc_esteid_card *estcard,
			 const struct sc_path *path,
			 const u8 *buf, size_t bufsize)
{
	char fname[160];
	int r;
        FILE *f;
        size_t c;

	r = generate_cache_filename(estcard, path, fname, sizeof(fname));
	if (r != 0)
		return r;
	f = fopen(fname, "w");
	if (f == NULL)
		return 0;
	c = fwrite(buf, 1, bufsize, f);
        fclose(f);
	if (c != bufsize) {
		error(estcard->card->ctx, "fwrite() wrote only %d bytes", c);
		unlink(fname);
		return SC_ERROR_INTERNAL;
	}
        return 0;
}

static struct sc_asn1_entry c_asn1_public_key[2] = {
	{ "publicKeyCoefficients", SC_ASN1_STRUCT, ASN1_SEQUENCE | SC_ASN1_CONS, },
	{ NULL }
};

static struct sc_asn1_entry c_asn1_rsa_pub_coefficients[3] = {
	{ "modulus",  SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC, },
	{ "exponent", SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC, },
	{ NULL }
};

static struct sc_asn1_entry c_asn1_dsa_pub_coefficients[5] = {
	{ "publicKey",SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC, },
	{ "paramP",   SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC, },
	{ "paramQ",   SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC, },
	{ "paramG",   SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC, },
	{ NULL },
};


int sc_esteid_decode_pubkey_dsa(struct sc_context *ctx,
		struct sc_esteid_pubkey_dsa *key,
		const u8 *buf, size_t buflen)
{
	struct sc_asn1_entry asn1_public_key[2];
	struct sc_asn1_entry asn1_dsa_pub_coeff[5];
	int r;
	
	debug(ctx,"decoding dsa pubkey\n");
	sc_copy_asn1_entry(c_asn1_public_key, asn1_public_key);
	sc_copy_asn1_entry(c_asn1_dsa_pub_coefficients, asn1_dsa_pub_coeff);

	sc_format_asn1_entry(asn1_public_key + 0, asn1_dsa_pub_coeff, NULL, 1);
	sc_format_asn1_entry(asn1_dsa_pub_coeff + 0,
				&key->pub.data, &key->pub.len, 0);
	sc_format_asn1_entry(asn1_dsa_pub_coeff + 1,
				&key->g.data, &key->g.len, 0);
	sc_format_asn1_entry(asn1_dsa_pub_coeff + 2,
				&key->p.data, &key->p.len, 0);
	sc_format_asn1_entry(asn1_dsa_pub_coeff + 3,
				&key->q.data, &key->q.len, 0);

	r = sc_asn1_decode(ctx, asn1_public_key, buf, buflen,
				NULL, NULL);
	SC_TEST_RET(ctx, r, "ASN.1 decoding failed");

	return 0;
}




int sc_esteid_decode_pubkey_rsa(struct sc_context *ctx,
	struct sc_esteid_pubkey_rsa *key,
	const u8 *buf, size_t buflen)
{
	struct sc_asn1_entry asn1_public_key[2];
	struct sc_asn1_entry asn1_rsa_coeff[3];
	int r;
	
	debug(ctx,"decoding rsa pubkey\n");
	sc_copy_asn1_entry(c_asn1_public_key, asn1_public_key);
	sc_format_asn1_entry(asn1_public_key + 0, asn1_rsa_coeff, NULL, 0);

	sc_copy_asn1_entry(c_asn1_rsa_pub_coefficients, asn1_rsa_coeff);
	sc_format_asn1_entry(asn1_rsa_coeff + 0,
				&key->modulus.data, &key->modulus.len, 0);
	sc_format_asn1_entry(asn1_rsa_coeff + 1,
				&key->exponent.data, &key->exponent.len, 0);

	r = sc_asn1_decode(ctx, asn1_public_key, buf, buflen, NULL, NULL);
	SC_TEST_RET(ctx, r, "ASN.1 parsing failed");

	return 0;
}

int sc_esteid_decode_pubkey(struct sc_context *ctx,
		struct sc_esteid_pubkey *key,
		const u8 *buf, size_t len)
{
	debug(ctx,"decoding pubkey\n");
	if (key->algorithm == SC_ALGORITHM_RSA)
		return sc_esteid_decode_pubkey_rsa(ctx, &key->u.rsa, buf, len);
	if (key->algorithm == SC_ALGORITHM_DSA)
		return sc_esteid_decode_pubkey_dsa(ctx, &key->u.dsa, buf, len);
	error(ctx, "Decoding of public key type %u not supported\n",
			key->algorithm);
	return SC_ERROR_NOT_SUPPORTED;
}

void sc_esteid_erase_pubkey(struct sc_esteid_pubkey *key)
{
	assert(key != NULL);
	switch (key->algorithm) {
	case SC_ALGORITHM_RSA:
		free(key->u.rsa.modulus.data);
		break;
	case SC_ALGORITHM_DSA:
		free(key->u.dsa.pub.data);
		free(key->u.dsa.g.data);
		free(key->u.dsa.p.data);
		free(key->u.dsa.q.data);
		break;
	}
	free(key->data.value);
	memset(key, 0, sizeof(*key));
}

static int parse_x509_cert(struct sc_context *ctx, const u8 *buf, size_t buflen, struct sc_esteid_cert *cert)
{
	int r;
	struct sc_esteid_pubkey *key = &cert->key;
	struct sc_algorithm_id pk_alg, sig_alg;
	sc_esteid_der_t pk = { NULL, 0 };
	struct sc_asn1_entry asn1_version[] = {
		{ "version",		SC_ASN1_INTEGER,   ASN1_INTEGER, 0, &cert->version },
		{ NULL }
	};
	struct sc_asn1_entry asn1_pkinfo[] = {
		{ "algorithm",		SC_ASN1_ALGORITHM_ID,  ASN1_SEQUENCE | SC_ASN1_CONS, 0, &pk_alg },
		{ "subjectPublicKey",	SC_ASN1_BIT_STRING_NI, ASN1_BIT_STRING, SC_ASN1_ALLOC, &pk.value, &pk.len },
		{ NULL }
	};
	struct sc_asn1_entry asn1_x509v3[] = {
		{ "certificatePolicies",	SC_ASN1_OCTET_STRING, SC_ASN1_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL },
		{ "subjectKeyIdentifier",	SC_ASN1_OCTET_STRING, SC_ASN1_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL },
		{ "crlDistributionPoints",	SC_ASN1_OCTET_STRING, SC_ASN1_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL | SC_ASN1_ALLOC, &cert->crl, &cert->crl_len },
		{ "authorityKeyIdentifier",	SC_ASN1_OCTET_STRING, SC_ASN1_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL },
		{ "keyUsage",			SC_ASN1_BOOLEAN, SC_ASN1_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL },
		{ NULL }
	};
	struct sc_asn1_entry asn1_extensions[] = {
		{ "x509v3",		SC_ASN1_STRUCT,    ASN1_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, asn1_x509v3 },
		{ NULL }
	};
	struct sc_asn1_entry asn1_tbscert[] = {
		{ "version",		SC_ASN1_STRUCT,    SC_ASN1_CTX | 0 | SC_ASN1_CONS, 0, asn1_version },
		{ "serialNumber",	SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC, &cert->serial, &cert->serial_len },
		{ "signature",		SC_ASN1_STRUCT,    ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL },
		{ "issuer",		SC_ASN1_OCTET_STRING, ASN1_SEQUENCE | SC_ASN1_CONS, SC_ASN1_ALLOC, &cert->issuer, &cert->issuer_len },
		{ "validity",		SC_ASN1_STRUCT,    ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL },
		{ "subject",		SC_ASN1_OCTET_STRING, ASN1_SEQUENCE | SC_ASN1_CONS, SC_ASN1_ALLOC, &cert->subject, &cert->subject_len },
		{ "subjectPublicKeyInfo",SC_ASN1_STRUCT,   ASN1_SEQUENCE | SC_ASN1_CONS, 0, asn1_pkinfo },
		{ "extensions",		SC_ASN1_STRUCT,    SC_ASN1_CTX | 3 | SC_ASN1_CONS, SC_ASN1_OPTIONAL, asn1_extensions },
		{ NULL }
	};
	struct sc_asn1_entry asn1_cert[] = {
		{ "tbsCertificate",	SC_ASN1_STRUCT,    ASN1_SEQUENCE | SC_ASN1_CONS, 0, asn1_tbscert },
		{ "signatureAlgorithm",	SC_ASN1_ALGORITHM_ID, ASN1_SEQUENCE | SC_ASN1_CONS, 0, &sig_alg },
		{ "signatureValue",	SC_ASN1_BIT_STRING,ASN1_BIT_STRING, 0, NULL, 0 },
		{ NULL }
	};
	const u8 *obj;
	size_t objlen;
	
	memset(cert, 0, sizeof(*cert));
	obj = sc_asn1_verify_tag(ctx, buf, buflen, ASN1_SEQUENCE | SC_ASN1_CONS,
				 &objlen);
	if (obj == NULL) {
		error(ctx, "X.509 certificate not found\n");
		return SC_ERROR_INVALID_ASN1_OBJECT;
	}
	cert->data_len = objlen + (obj - buf);
	r = sc_asn1_decode(ctx, asn1_cert, obj, objlen, NULL, NULL);
	SC_TEST_RET(ctx, r, "ASN.1 parsing failed");

	cert->version++;

	key->algorithm = pk_alg.algorithm;
	pk.len >>= 3;	/* convert number of bits to bytes */
	key->data = pk;

	r= sc_esteid_decode_pubkey(ctx,key,pk.value, pk.len);

	if (r < 0)
		free(pk.value);
	sc_asn1_clear_algorithm_id(&pk_alg);
	sc_asn1_clear_algorithm_id(&sig_alg);

	return r;
}


int sc_esteid_read_certificate(struct sc_esteid_card *estcard,
			       const struct sc_esteid_cert_info *info,
			       struct sc_esteid_cert **cert_out)
{
	int r;
	struct sc_esteid_cert *cert;
	struct sc_file *file = NULL;
	u8 *data = NULL;
	size_t len;
	
	assert(estcard != NULL && info != NULL && cert_out != NULL);
	SC_FUNC_CALLED(estcard->card->ctx, 1);
	r = sc_esteid_read_cached_file(estcard, &info->path, &data, &len);
	if (r) {
		r = select_esteid_df( estcard);
		SC_TEST_RET(estcard->card->ctx, r, "select_esteid_df() failed");
		r = sc_lock(estcard->card);
		SC_TEST_RET(estcard->card->ctx, r, "sc_lock() failed");
		r = sc_select_file(estcard->card, &info->path, &file);
		if (r) {
			sc_unlock(estcard->card);
			return r;
		}
		len = file->size;
		sc_file_free(file);
		data = (u8 *) malloc(len);
		if (data == NULL) {
			sc_unlock(estcard->card);
			return SC_ERROR_OUT_OF_MEMORY;
		}
		r = sc_read_binary(estcard->card, 0, data, len, 0);
		if (r < 0) {
			sc_unlock(estcard->card);
			free(data);
			return r;
		}
		sc_unlock(estcard->card);
	}
	cert = (struct sc_esteid_cert *) malloc(sizeof(struct sc_esteid_cert));
	if (cert == NULL) {
		free(data);
		return SC_ERROR_OUT_OF_MEMORY;
	}
	memset(cert, 0, sizeof(struct sc_esteid_cert));
	if (parse_x509_cert(estcard->card->ctx, data, len, cert)) {
		free(data);
		free(cert);
		return SC_ERROR_INVALID_ASN1_OBJECT;
	}
	cert->data = data;
	*cert_out = cert;
	return 0;
}

void sc_esteid_free_certificate(struct sc_esteid_cert *cert)
{
	assert(cert != NULL);

	sc_esteid_erase_pubkey(&cert->key);
	free(cert->subject);
	free(cert->issuer);
	free(cert->serial);
	free(cert->data);
	free(cert);
}


int sc_esteid_verify_pin(struct sc_esteid_card *estcard,
			 struct sc_esteid_pin_info *pin,
			 const u8 *pincode, size_t pinlen)
{
	int r;
	struct sc_card *card;
	u8 pinbuf[SC_MAX_PIN_SIZE];

	assert(estcard != NULL);
	if (pin->magic != SC_ESTEID_PIN_MAGIC)
		return SC_ERROR_OBJECT_NOT_VALID;
	if (pinlen > pin->stored_length || pinlen < pin->min_length)
		return SC_ERROR_INVALID_PIN_LENGTH;
	card = estcard->card;
	r = sc_lock(card);
	SC_TEST_RET(card->ctx, r, "sc_lock() failed");
	memset(pinbuf, pin->pad_char, pin->stored_length);
	memcpy(pinbuf, pincode, pinlen);
	pinbuf[pinlen]='\000';

	r = sc_verify(card, SC_AC_CHV, pin->reference,pinbuf, pinlen, &pin->tries_left);
	
	memset(pinbuf, 0, pinlen);
	sc_unlock(card);
	if (r)
		return r;

	return 0;
}

int sc_esteid_change_pin(struct sc_esteid_card *estcard,
			 struct sc_esteid_pin_info *pin,
			 const u8 *oldpin, size_t oldpinlen,
			 const u8 *newpin, size_t newpinlen)
{
	int r;
	struct sc_card *card;

	assert(estcard != NULL);
	if (pin->magic != SC_ESTEID_PIN_MAGIC)
		return SC_ERROR_OBJECT_NOT_VALID;
	if ((oldpinlen > pin->stored_length)
	    || (newpinlen > pin->stored_length))
		return SC_ERROR_INVALID_ARGUMENTS;
	if ((oldpinlen < pin->min_length) || (newpinlen < pin->min_length))
		return SC_ERROR_INVALID_ARGUMENTS;
	card = estcard->card;
	r = sc_lock(card);
	SC_TEST_RET(card->ctx, r, "sc_lock() failed");

	r = sc_change_reference_data(card, SC_AC_CHV,pin->reference, oldpin, 
		oldpinlen, newpin, newpinlen, &pin->tries_left);

	sc_unlock(card);
	return r;
}

int sc_esteid_unblock_pin(struct sc_esteid_card *estcard,
			 struct sc_esteid_pin_info *puk,
			 const u8 *pukcode, size_t puklen, int ref)
{
	int r;
	struct sc_card *card;

	assert(estcard != NULL);
	if (puk->magic != SC_ESTEID_PIN_MAGIC)
		return SC_ERROR_OBJECT_NOT_VALID;
	if (puklen > puk->stored_length)
		return SC_ERROR_INVALID_ARGUMENTS;
	if (puklen < puk->min_length)
		return SC_ERROR_INVALID_ARGUMENTS;
	card = estcard->card;
	r = sc_lock(card);
	SC_TEST_RET(card->ctx, r, "sc_lock() failed");

	r = sc_esteid_verify_pin(estcard, puk, pukcode, puklen);
	r = sc_reset_retry_counter(card, SC_AC_CHV, ref, NULL, 0, NULL, 0);
	sc_unlock(card);
	return r;
}



int sc_esteid_unblock_and_change_pin(struct sc_esteid_card *estcard,
			 struct sc_esteid_pin_info *puk,
			 const u8 *pukcode, size_t puklen,
			 struct sc_esteid_pin_info *newpin,
			 const u8 *newpincode, size_t newpinlen)
{
	int r;
	struct sc_card *card;

	assert(estcard != NULL);
	if ((puk->magic != SC_ESTEID_PIN_MAGIC)
		|| (newpin->magic!= SC_ESTEID_PIN_MAGIC))
		return SC_ERROR_OBJECT_NOT_VALID;
	if ((puklen > puk->stored_length)
	    || (newpinlen > newpin->stored_length))
		return SC_ERROR_INVALID_ARGUMENTS;
	if ((puklen < puk->min_length) || (newpinlen < newpin->min_length))
		return SC_ERROR_INVALID_ARGUMENTS;
	card = estcard->card;
	r = sc_lock(card);
	SC_TEST_RET(card->ctx, r, "sc_lock() failed");

	r=sc_reset_retry_counter(card, SC_AC_CHV, newpin->reference,
		pukcode, puklen, newpincode, newpinlen);
	sc_unlock(card);
	return r;
}



int sc_esteid_get_key_usage(struct sc_esteid_card *estcard, int rec)
{
	int r;
	u8 databuf[SC_MAX_APDU_BUFFER_SIZE-3];	//fixme -- mf
	struct sc_path path;

	//rec:
	//sign ver 1 = 1
	//sign ver 2 = 2
	//auth ver 1 = 3
	//auth ver 2 = 4


	debug(estcard->card->ctx, "called\n");	

	sc_format_path("3F00", &path);
	path.type = SC_PATH_TYPE_FILE_ID;
	r = sc_select_file( estcard->card, &path, NULL);
	if (r) {
		error(estcard->card->ctx, "Error selecting EstEID MF(DIR): %s\n", sc_strerror(r));
		return 0;
	}
	sc_format_path("EEEE", &path);
	path.type = SC_PATH_TYPE_FILE_ID;
	r = sc_select_file( estcard->card, &path, NULL);
	if (r) {
		error(estcard->card->ctx, "Error selecting file  'EEEE': %s\n", sc_strerror(r));
		return 0;
	}
	sc_format_path("0013", &path);
	path.type = SC_PATH_TYPE_FILE_ID;
	r = sc_select_file( estcard->card, &path, NULL);
	if (r) {
		error(estcard->card->ctx, "Error selecting file  '0013': %s\n", sc_strerror(r));
		return 0;
	}

	r = sc_read_record( estcard->card, rec, databuf, sizeof(databuf),  0);
		if (r < 0) {
			error(estcard->card->ctx, "READ RECORD from '0013' failed: %s\n", sc_strerror(r));
			return 0;
		}

	r = 0xffffff - (databuf[12] * 65536 + databuf[13] * 256 + databuf[14]);
	return r;
}


