/*
 * framework-esteid.c: EstEID framework and related objects
 *
 * Copyright (C) 2002  Timo Ters <timo.teras@iki.fi>
 * Copyright (C) 2002  Priit Randla <priit.randla@eyp.ee>
 * Copyright (C) 2002  Marie Fischer <marie@vtl.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include "sc-pkcs11.h"
#include "../libopensc/esteid.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define MAX_CACHE_PIN		32
struct esteid_slot_data {
	struct sc_esteid_object *auth_obj;
	struct {
		u8		value[MAX_CACHE_PIN];
		unsigned int	len;
	}			
	pin[2];
};
#define slot_data(p)		((struct esteid_slot_data *) p)
#define slot_data_auth(p)	(slot_data(p)->auth_obj)
#define slot_data_pin_info(p)	(slot_data_auth(p)? \
		(struct sc_esteid_pin_info *) slot_data_auth(p)->data : NULL)

#define check_attribute_buffer(attr,size)	\
	if (attr->pValue == NULL_PTR) {         \
		attr->ulValueLen = size;        \
		return CKR_OK;                  \
	}                                       \
	if (attr->ulValueLen < size) {		\
		attr->ulValueLen = size;	\
		return CKR_BUFFER_TOO_SMALL;    \
	}                                       \
        attr->ulValueLen = size;

extern struct sc_pkcs11_object_ops esteid_cert_ops;
extern struct sc_pkcs11_object_ops esteid_prkey_ops;
extern struct sc_pkcs11_object_ops esteid_cert_key_ops;
extern struct sc_pkcs11_object_ops esteid_pubkey_ops;

struct esteid_cert_object {
	struct sc_pkcs11_object object;
	struct sc_esteid_object *certificate_object;
	struct sc_esteid_cert_info *certificate_info;
	struct sc_esteid_cert *certificate;
};

struct esteid_prkey_object {
	struct sc_pkcs11_object object;
	struct sc_esteid_object *prkey_object;
	struct sc_esteid_prkey_info *prkey_info;
	struct esteid_cert_object *cert_object;
	struct esteid_pubkey_object *pubkey_object;
};

struct esteid_cert_key_object {
	struct sc_pkcs11_object object;
	struct sc_esteid_object *certificate_object;
	struct sc_esteid_cert_info *certificate_info;
	struct sc_esteid_pubkey *key;
};


struct esteid_pubkey_object {
	struct sc_pkcs11_object object;
	struct sc_esteid_object *pubkey_object;
	struct sc_esteid_pubkey_info *pubkey_info;
	struct sc_esteid_pubkey *key;
};


static int	get_public_exponent(struct sc_esteid_pubkey *,
					CK_ATTRIBUTE_PTR);
static int	get_modulus(struct sc_esteid_pubkey *,
					CK_ATTRIBUTE_PTR);
static int	get_modulus_bits(struct sc_esteid_pubkey *,
					CK_ATTRIBUTE_PTR);
static int	asn1_sequence_wrapper(const u8 *, size_t, CK_ATTRIBUTE_PTR);
static void	cache_pin(void *, int, const void *, size_t);

/* EstEID Framework */

static CK_RV esteid_bind(struct sc_pkcs11_card *p11card)
{
	int rc = sc_esteid_bind(p11card->card, (struct sc_esteid_card**) &p11card->fw_data);
	debug(context, "Binding to EstEID, rc=%d\n", rc);
	return sc_to_cryptoki_error(rc, p11card->reader);
}

static CK_RV esteid_unbind(struct sc_pkcs11_card *p11card)
{
	struct sc_esteid_card *card = (struct sc_esteid_card*) p11card->fw_data;
	int rc = sc_esteid_unbind(card);
	return sc_to_cryptoki_error(rc, p11card->reader);
}

static void esteid_init_token_info(struct sc_esteid_card *card, CK_TOKEN_INFO_PTR pToken)
{
	strcpy_bp(pToken->manufacturerID, card->manufacturer_id, 32);
	strcpy_bp(pToken->model, "EstEID Card", 16);
	strcpy_bp(pToken->serialNumber, card->serial_number, 16);
	pToken->ulMaxSessionCount = CK_EFFECTIVELY_INFINITE;
	pToken->ulSessionCount = 0; /* FIXME */
	pToken->ulMaxRwSessionCount = CK_EFFECTIVELY_INFINITE;
	pToken->ulRwSessionCount = 0; /* FIXME */
	pToken->ulTotalPublicMemory = CK_UNAVAILABLE_INFORMATION;
	pToken->ulFreePublicMemory = CK_UNAVAILABLE_INFORMATION;
	pToken->ulTotalPrivateMemory = CK_UNAVAILABLE_INFORMATION;
	pToken->ulFreePrivateMemory = CK_UNAVAILABLE_INFORMATION;
	pToken->hardwareVersion.major = 1;
	pToken->hardwareVersion.minor = 0;
	pToken->firmwareVersion.major = 1;
	pToken->firmwareVersion.minor = 0;
}

static struct esteid_cert_object *
esteid_add_cert_object(struct sc_pkcs11_slot *slot,
		       struct sc_esteid_card *card,
		       struct sc_esteid_object *cert,
		       CK_OBJECT_HANDLE_PTR pHandle)
{
	struct esteid_cert_object *object;
	struct esteid_cert_key_object *obj2;

	/* Certificate object */
	object = (struct esteid_cert_object*) calloc(1, sizeof(struct esteid_cert_object));
	object->object.ops = &esteid_cert_ops;
	object->certificate_object = cert;
	object->certificate_info = (struct sc_esteid_cert_info*) cert->data;
	sc_esteid_read_certificate(card, object->certificate_info, &object->certificate);
	pool_insert(&slot->object_pool, object, pHandle);

	/* Corresponding public key */
	obj2 = (struct esteid_cert_key_object*) calloc(1, sizeof(struct esteid_cert_key_object));
	obj2->object.ops = &esteid_cert_key_ops;
	obj2->key = &object->certificate->key;
	obj2->certificate_object = cert;
	obj2->certificate_info = (struct sc_esteid_cert_info*) cert->data;
	pool_insert(&slot->object_pool, obj2, NULL);

	/* Mark as seen */
	cert->flags |= SC_ESTEID_CO_FLAG_OBJECT_SEEN;

	return object;
}


static struct esteid_prkey_object *
esteid_add_prkey_object(struct sc_pkcs11_slot *slot,
			struct sc_esteid_card *card,
			struct sc_esteid_object *prkey,
			struct sc_esteid_object **certs,
			int cert_count,
			struct sc_esteid_object **pubkeys,
			int pubkey_count,
			CK_OBJECT_HANDLE_PTR pHandle
			)
{
	struct esteid_prkey_object *object;
	int i;

	struct sc_esteid_cert *obj1;
	struct esteid_pubkey_object *obj2;
	
	object = (struct esteid_prkey_object*) calloc(1, sizeof(struct esteid_prkey_object));
	object->object.ops = &esteid_prkey_ops;
	object->prkey_object = prkey;
	object->prkey_info = (struct sc_esteid_prkey_info*) prkey->data;
	pool_insert(&slot->object_pool, object, pHandle);

	/* Mark as seen */
	prkey->flags |= SC_ESTEID_CO_FLAG_OBJECT_SEEN;

	/* Also add the related certificate if found */
	for (i=0; i < cert_count; i++) {
		if (sc_esteid_compare_id(&object->prkey_info->id, &((struct sc_esteid_cert_info*)certs[i]->data)->id)) {
			debug(context, "Adding certificate %d relating to private key\n", i);
			object->cert_object = esteid_add_cert_object(slot, card, certs[i], NULL);
			break;
		}
	}

	return object;
}

static void esteid_init_slot(struct sc_esteid_card *card,
		struct sc_pkcs11_slot *slot,
		struct sc_esteid_object *auth)
{
	struct esteid_slot_data *fw_data;
	struct sc_esteid_pin_info *pin_info = NULL;
	char tmp[64];
	esteid_init_token_info(card, &slot->token_info);
	slot->token_info.flags = CKF_USER_PIN_INITIALIZED
				| CKF_TOKEN_INITIALIZED
				| CKF_WRITE_PROTECTED;
	slot->fw_data = fw_data = (struct esteid_slot_data *) calloc(1, sizeof(*fw_data));
	fw_data->auth_obj = auth;
	if (auth != NULL) {
		pin_info = (struct sc_esteid_pin_info*) auth->data;
		snprintf(tmp, sizeof(tmp), "%s (%s)",card->label, auth->label);
		slot->token_info.flags |= CKF_LOGIN_REQUIRED;
	} else{
		sprintf(tmp, "public");
	}
	strcpy_bp(slot->token_info.label, tmp, 32);
	if (pin_info && pin_info->magic == SC_ESTEID_PIN_MAGIC) {
		slot->token_info.ulMaxPinLen = pin_info->stored_length;
		slot->token_info.ulMinPinLen = pin_info->min_length;
	} else {
		/* choose reasonable defaults */
		slot->token_info.ulMaxPinLen = 8;
		slot->token_info.ulMinPinLen = 4;
	}

	debug(context, "Initialized token '%s'\n", tmp);
}

static CK_RV esteid_create_slot(struct sc_pkcs11_card *p11card,
		struct sc_esteid_object *auth,
		struct sc_pkcs11_slot **out)
{
	struct sc_esteid_card *card = (struct sc_esteid_card*) p11card->fw_data;
	int rv;

	rv = slot_allocate(out, p11card);
	if (rv != CKR_OK)
		return rv;

	esteid_init_slot(card, *out, auth);
	return CKR_OK;
}

static CK_RV esteid_create_tokens(struct sc_pkcs11_card *p11card)
{
	struct sc_esteid_card *card = (struct sc_esteid_card*) p11card->fw_data;
	struct sc_pkcs11_slot *slot;

	int app,i, j, rv, reader = p11card->reader;
	int auth_count, cert_count, prkey_count, pubkey_count;

	pubkey_count=0;
	app=SC_ESTEID_APP_PKCS11;
	
	rv=cert_count = sc_esteid_enum_certificates (card, app);
	if (rv < 0)
		return sc_to_cryptoki_error(rv, reader);
	debug(context, "Found %d certs\n", cert_count);

	rv=prkey_count = sc_esteid_enum_private_keys (card, app);
	if (rv < 0)
		return sc_to_cryptoki_error(rv, reader);
	debug(context, "Found %d prkeys\n", prkey_count);

	rv=auth_count = sc_esteid_enum_pins (card, app);
	if (rv < 0)
		return sc_to_cryptoki_error(rv, reader);
	debug(context, "Found %d pins\n", auth_count);

	for (i = 0; i < auth_count; i++) {

		struct sc_esteid_pin_info *pin_info = NULL;

		pin_info = (struct sc_esteid_pin_info*) card->pins[i]->data;

		/* Ignore any non-authentication PINs */
		if ((pin_info->flags & SC_ESTEID_PIN_FLAG_SO_PIN)
		 || (pin_info->flags & SC_ESTEID_PIN_FLAG_UNBLOCKING_PIN))
			continue;

		/* Add all the private keys related to this pin */
		rv = esteid_create_slot(p11card, card->pins[i], &slot);

		if (rv != CKR_OK)
			return rv;

		for (j=0; j < prkey_count; j++) {

			if (sc_esteid_compare_id(&pin_info->auth_id,&card->prkeys[j]->auth_id)) {
				debug(context, "Adding private key %d to PIN %d\n", j, i);
				esteid_add_prkey_object(slot, card, card->prkeys[j],card->certs, cert_count,NULL, 0,NULL);
			}
		}
	}

	slot=NULL;

	for (i=0; i < cert_count; i++) {
		// XXX netscape wants to see all certificates in a slot
		// that doesn't require login
		// doesn't work for me -- mf
		if (!(card->certs[i]->flags & SC_ESTEID_CO_FLAG_OBJECT_SEEN)) {
            debug(context, "Certificate %d was not seen previously\n", i);
			if (!slot) {
				rv = esteid_create_slot(p11card, NULL, &slot);
				if (rv != CKR_OK)
					return rv;
			}
			esteid_add_cert_object(slot, card, card->certs[i], NULL);
		}
	}

	/* Create read/write slots */
//	don't need those i guess -- mf
//	while (slot_allocate(&slot, p11card) == CKR_OK) {
//		esteid_init_token_info(card, &slot->token_info);
//		slot->token_info.flags = CKF_TOKEN_INITIALIZED;
//	}

	debug(context, "All tokens created\n");
	return CKR_OK;
}

//from framework-pkcs15.c -- mf
static CK_RV esteid_release_token(struct sc_pkcs11_card *p11card, void *fw_token)
{
        return CKR_OK;
}

//from framework-pkcs15.c -- mf
static CK_RV esteid_get_mechanism_list(struct sc_pkcs11_card *p11card,
				       void *fw_token,
				       CK_MECHANISM_TYPE_PTR pMechanismList,
				       CK_ULONG_PTR pulCount)
{
	static const CK_MECHANISM_TYPE mechanism_list[] = {
		CKM_RSA_PKCS,
		CKM_SHA1_RSA_PKCS,
		CKM_SHA_1,
	};
	const int numMechanisms = sizeof(mechanism_list) / sizeof(mechanism_list[0]);

	if (pMechanismList == NULL_PTR) {
		*pulCount = numMechanisms;
		return CKR_OK;
	}

	if (*pulCount < numMechanisms) {
		*pulCount = numMechanisms;
		return CKR_BUFFER_TOO_SMALL;
	}

	memcpy(pMechanismList, &mechanism_list, sizeof(mechanism_list));

	return CKR_OK;
}

//from framework-pkcs15.c -- mf
static CK_RV esteid_get_mechanism_info(struct sc_pkcs11_card *p11card,
				       void *fw_token,
				       CK_MECHANISM_TYPE type,
				       CK_MECHANISM_INFO_PTR pInfo)
{
	switch (type) {
	case CKM_RSA_PKCS:
	case CKM_SHA1_RSA_PKCS:
		pInfo->flags = CKF_HW | CKF_SIGN | CKF_DECRYPT;
		pInfo->ulMinKeySize = 512;
		pInfo->ulMaxKeySize = 2048;
		break;
	case CKM_SHA_1:
                pInfo->flags = CKF_HW | CKF_DIGEST;
                pInfo->ulMinKeySize = 0;      /* These are meaningless here */
                pInfo->ulMaxKeySize = 0;
                break;

	default:
		return CKR_MECHANISM_INVALID;
	}
	return CKR_OK;
}

static CK_RV esteid_login(struct sc_pkcs11_card *p11card,
			  void *fw_token,
			  CK_USER_TYPE userType,
			  CK_CHAR_PTR pPin,
			  CK_ULONG ulPinLen)
{
	int rc;
	struct sc_esteid_card *card = (struct sc_esteid_card*) p11card->fw_data;
	struct sc_esteid_object *auth_object;
	struct sc_esteid_pin_info *pin;

	switch (userType) {
	case CKU_USER:
		auth_object = slot_data_auth(fw_token);
		if (auth_object == NULL)
			return CKR_USER_PIN_NOT_INITIALIZED;
		break;
	default:
		return CKR_USER_TYPE_INVALID;
	}
	pin = (struct sc_esteid_pin_info *) auth_object->data;
	if (ulPinLen < pin->min_length || ulPinLen > pin->stored_length){
		return CKR_PIN_LEN_RANGE;
	}
	rc = sc_lock(card->card);
	if (rc < 0) {
		 debug(context, "Failed to lock card (%d)\n", rc);
		 return sc_to_cryptoki_error(rc, p11card->reader);
	}

	rc = sc_esteid_verify_pin(card, pin, pPin, ulPinLen);
	debug(context, "PIN verification returned %d\n", rc);
	
	if (rc >= 0)
		cache_pin(fw_token, userType, pPin, ulPinLen);

	return sc_to_cryptoki_error(rc, p11card->reader);
}

//from framework-pkcs15.c -- mf
static CK_RV esteid_logout(struct sc_pkcs11_card *p11card, void *fw_token)
{
	struct sc_esteid_card *card = (struct sc_esteid_card*) p11card->fw_data;
	int rc;

	cache_pin(fw_token, CKU_SO, NULL, 0);
	cache_pin(fw_token, CKU_USER, NULL, 0);

	rc = sc_unlock(card->card);
	return sc_to_cryptoki_error(rc, p11card->reader);
}

//from framework-pkcs15.c, untested -- mf
static CK_RV esteid_change_pin(struct sc_pkcs11_card *p11card,
			  void *fw_token,
			  CK_CHAR_PTR pOldPin, CK_ULONG ulOldLen,
			  CK_CHAR_PTR pNewPin, CK_ULONG ulNewLen)
{
	int rc;
	struct sc_esteid_card *card = (struct sc_esteid_card*) p11card->fw_data;
	struct sc_esteid_pin_info *pin;

	if (!(pin = slot_data_pin_info(fw_token)))
		return CKR_USER_PIN_NOT_INITIALIZED;

	if (ulNewLen < pin->min_length ||
	    ulNewLen > pin->stored_length)
		return CKR_PIN_LEN_RANGE;

	rc = sc_esteid_change_pin(card, pin, pOldPin, ulOldLen,
				pNewPin, ulNewLen);
        debug(context, "PIN verification returned %d\n", rc);

	if (rc >= 0)
		cache_pin(fw_token, CKU_USER, pNewPin, ulNewLen);
	return sc_to_cryptoki_error(rc, p11card->reader);
}

struct sc_pkcs11_framework_ops framework_esteid = {
	esteid_bind,
	esteid_unbind,
	esteid_create_tokens,
	esteid_release_token,
	esteid_get_mechanism_list,
	esteid_get_mechanism_info,
	esteid_login,
	esteid_logout,
	esteid_change_pin,
	NULL,			/* init_token */
	NULL,
	NULL
};

/*
 * EstEID Certificate Object
 */


//from framework-pkcs15.c -- mf
void esteid_cert_release(void *obj)
{
	struct esteid_cert_object *cert = (struct esteid_cert_object *) obj;
	sc_esteid_free_certificate(cert->certificate);
}

//from framework-pkcs15.c -- mf
CK_RV esteid_cert_get_attribute(struct sc_pkcs11_session *session,
				void *object,
				CK_ATTRIBUTE_PTR attr)
{
	struct esteid_cert_object *cert = (struct esteid_cert_object*) object;
	debug(context,"called\n");	
	switch (attr->type) {
	case CKA_CLASS:
		debug(context, "class\n");
		check_attribute_buffer(attr, sizeof(CK_OBJECT_CLASS));
		*(CK_OBJECT_CLASS*)attr->pValue = CKO_CERTIFICATE;
		break;
	case CKA_TOKEN:
		debug(context, "token\n");
		check_attribute_buffer(attr, sizeof(CK_BBOOL));
		*(CK_BBOOL*)attr->pValue = TRUE;
		break;
	case CKA_PRIVATE:
	case CKA_MODIFIABLE:
		debug(context, "private/modifiable\n");
		check_attribute_buffer(attr, sizeof(CK_BBOOL));
		*(CK_BBOOL*)attr->pValue = FALSE;
		break;
	case CKA_LABEL:
		debug(context, "label\n");
		check_attribute_buffer(attr, strlen(cert->certificate_object->label));
		memcpy(attr->pValue, cert->certificate_object->label, strlen(cert->certificate_object->label));
		break;
	case CKA_CERTIFICATE_TYPE:
		debug(context, "type\n");
		check_attribute_buffer(attr, sizeof(CK_CERTIFICATE_TYPE));
		*(CK_CERTIFICATE_TYPE*)attr->pValue = CKC_X_509;
		break;
	case CKA_ID:
		debug(context, "id\n");
		if (cert->certificate_info->authority) {
			check_attribute_buffer(attr, 1);
			*(unsigned char*)attr->pValue = 0;
		} else {
			check_attribute_buffer(attr, cert->certificate_info->id.len);
			memcpy(attr->pValue, cert->certificate_info->id.value, cert->certificate_info->id.len);
		}
		break;
	case CKA_TRUSTED:
		debug(context, "trusted\n");
		check_attribute_buffer(attr, sizeof(CK_BBOOL));
		*(CK_BBOOL*)attr->pValue = cert->certificate_info->authority?TRUE:FALSE;
		break;
	case CKA_VALUE:
		debug(context, "value\n");
		check_attribute_buffer(attr, cert->certificate->data_len);
		memcpy(attr->pValue, cert->certificate->data, cert->certificate->data_len);
		break;
	case CKA_SERIAL_NUMBER:
		debug(context, "serial\n");
		check_attribute_buffer(attr, cert->certificate->serial_len);
		memcpy(attr->pValue, cert->certificate->serial, cert->certificate->serial_len);
		break;
	case CKA_SUBJECT:
		debug(context, "subject\n");
		return asn1_sequence_wrapper(cert->certificate->subject,
			cert->certificate->subject_len,
			attr);
	case CKA_ISSUER:
		debug(context, "issuer\n");
		return asn1_sequence_wrapper(cert->certificate->issuer,
			cert->certificate->issuer_len,
			attr);
	default:
		debug(context, "invalid\n");
		return CKR_ATTRIBUTE_TYPE_INVALID;
	}
	return CKR_OK;
}

//from framework-pkcs15.c -- mf
static int
esteid_cert_cmp_attribute(struct sc_pkcs11_session *session,
				void *object,
				CK_ATTRIBUTE_PTR attr)
{
	struct esteid_cert_object *cert = (struct esteid_cert_object*) object;
	u8	*data;
	size_t	len;

	switch (attr->type) {
	/* Check the issuer. Some pkcs11 callers (i.e. netscape) will pass
	 * in the ASN.1 encoded SEQUENCE OF SET ... while OpenSC just
	 * keeps the SET in the issuer field. */
	case CKA_ISSUER:
		if (cert->certificate->issuer_len == 0)
			break;
		data = (u8 *) attr->pValue;
		len = attr->ulValueLen;
		/* SEQUENCE is tag 0x30, SET is 0x31
		 * I know this code is icky, but hey... this is netscape
		 * we're dealing with :-) */
		if (cert->certificate->issuer[0] == 0x31
		 && data[0] == 0x30 && len >= 2) {
			/* skip the length byte(s) */
			len = (data[1] & 0x80)? (data[1] & 0x7F) : 0;
			if (attr->ulValueLen < len + 2)
				break;
			data += len + 2;
			len = attr->ulValueLen - len - 2;
		}
		if (len == cert->certificate->issuer_len
		 && !memcmp(cert->certificate->issuer, data, len))
			return 1;
		break;
	default:
		return sc_pkcs11_any_cmp_attribute(session, object, attr);
	}

	return 0;
}

struct sc_pkcs11_object_ops esteid_cert_ops = {
	esteid_cert_release,
	NULL,
	esteid_cert_get_attribute,
	esteid_cert_cmp_attribute,
	NULL,
	NULL,
	NULL
};

/*
 * EstEID Private Key Object
 */

//from framework-pkcs15.c -- mf
CK_RV esteid_prkey_get_attribute(struct sc_pkcs11_session *session,
				void *object,
				CK_ATTRIBUTE_PTR attr)
{
	struct esteid_prkey_object *prkey = (struct esteid_prkey_object*) object;
	struct sc_esteid_pubkey *key = NULL;
	debug(context,"called\n");
	if (prkey->cert_object && prkey->cert_object->certificate)
		key = &prkey->cert_object->certificate->key;
	else if (prkey->pubkey_object)
		key = prkey->pubkey_object->key;

	switch (attr->type) {
	case CKA_CLASS:
		check_attribute_buffer(attr, sizeof(CK_OBJECT_CLASS));
		*(CK_OBJECT_CLASS*)attr->pValue = CKO_PRIVATE_KEY;
                break;
	case CKA_TOKEN:
	case CKA_LOCAL:
	case CKA_SENSITIVE:
	case CKA_ALWAYS_SENSITIVE:
	case CKA_NEVER_EXTRACTABLE:
	case CKA_SIGN:
	case CKA_PRIVATE:
	case CKA_UNWRAP: /* XXX should make an attempt to find out whether
			    the card supports unwrap */
		check_attribute_buffer(attr, sizeof(CK_BBOOL));
		*(CK_BBOOL*)attr->pValue = TRUE;
		break;
	case CKA_MODIFIABLE:
	case CKA_DERIVE:
	case CKA_DECRYPT:
	case CKA_SIGN_RECOVER:
	case CKA_EXTRACTABLE:
		check_attribute_buffer(attr, sizeof(CK_BBOOL));
		*(CK_BBOOL*)attr->pValue = FALSE;
		break;
	case CKA_LABEL:
		check_attribute_buffer(attr, strlen(prkey->prkey_object->label));
		memcpy(attr->pValue, prkey->prkey_object->label, strlen(prkey->prkey_object->label));
		break;
	case CKA_KEY_TYPE:
		check_attribute_buffer(attr, sizeof(CK_KEY_TYPE));
		*(CK_KEY_TYPE*)attr->pValue = CKK_RSA;
		break;
	case CKA_ID:
		check_attribute_buffer(attr, prkey->prkey_info->id.len);
		memcpy(attr->pValue, prkey->prkey_info->id.value, prkey->prkey_info->id.len);
		break;
	case CKA_KEY_GEN_MECHANISM:
		check_attribute_buffer(attr, sizeof(CK_MECHANISM_TYPE));
		*(CK_MECHANISM_TYPE*)attr->pValue = CK_UNAVAILABLE_INFORMATION;
		break;
	case CKA_MODULUS:
		return get_modulus(key, attr);
	case CKA_MODULUS_BITS:
		return get_modulus_bits(key, attr);
	case CKA_PUBLIC_EXPONENT:
		return get_public_exponent(key, attr);
	case CKA_PRIVATE_EXPONENT:
	case CKA_PRIME_1:
	case CKA_PRIME_2:
	case CKA_EXPONENT_1:
	case CKA_EXPONENT_2:
	case CKA_COEFFICIENT:
		return CKR_ATTRIBUTE_SENSITIVE;
		/*
		 case CKA_SUBJECT:
		 case CKA_START_DATE:
		 case CKA_END_DATE:
		 */
	default:
		return CKR_ATTRIBUTE_TYPE_INVALID;
	}
	return CKR_OK;
}

//from framework-pkcs15.c -- mf
CK_RV esteid_prkey_sign(struct sc_pkcs11_session *ses, void *obj,
			CK_MECHANISM_PTR pMechanism, CK_BYTE_PTR pData,
			CK_ULONG ulDataLen, CK_BYTE_PTR pSignature,
			CK_ULONG_PTR pulDataLen)
{
	struct esteid_prkey_object *prkey = (struct esteid_prkey_object *) obj;
	int rv, flags = 0;

	debug(context, "Initiating signing operation.\n");

	flags = SC_ALGORITHM_RSA_PAD_PKCS1;
	switch (pMechanism->mechanism) {
	case CKM_RSA_PKCS:
		/* Um. We need to guess what netscape is trying to
		 * sign here. We're lucky that all these things have
		 * different sizes. */
		switch (ulDataLen) {
		case 34:flags |= SC_ALGORITHM_RSA_HASH_MD5;  /* MD5 + header */
			pData += 18; ulDataLen -= 18;
			debug(context,"34\n");
			break;
		case 35:flags |= SC_ALGORITHM_RSA_HASH_SHA1;   /* SHA1 + hdr */
//fixme - need to do this nicer ...
// what algorithms do we do, anyway?
//			pData += 15; ulDataLen -= 15;
			debug(context,"35\n");
			break;
		case 36:flags |= SC_ALGORITHM_RSA_HASH_MD5_SHA1; /* SSL hash */
			debug(context,"36\n");
			break;
		case 20:
			flags |= SC_ALGORITHM_RSA_HASH_SHA1;	/* SHA1 */
			debug(context,"20\n");
			break;
		case 16:
			flags |= SC_ALGORITHM_RSA_HASH_MD5;	/* MD5 */
			debug(context,"16\n");
			break;
		default:
			debug(context,"default\n");
			flags |= SC_ALGORITHM_RSA_HASH_NONE;
		}
		break;
	case CKM_SHA1_RSA_PKCS:
		flags |= SC_ALGORITHM_RSA_HASH_SHA1;
		debug(context,"rsa hash sha1\n");
		break;
	default:
                return CKR_MECHANISM_INVALID;
	}

	debug(context, "Selected flags %X. Now computing signature for %d bytes. %d bytes reserved.\n", flags, ulDataLen, *pulDataLen);
	
	//orig was prkey->prkey_object instead of prkey->prkey_info -- mf
	rv = sc_esteid_compute_signature((struct sc_esteid_card*) ses->slot->card->fw_data,
					 prkey->prkey_info,
					 flags,
					 pData,
					 ulDataLen,
					 pSignature,
					 *pulDataLen);
	
	debug(context, "Sign complete. Result %d.\n", rv);

	if (rv > 0) {
		return CKR_OK;
	}
	
	return sc_to_cryptoki_error(rv, ses->slot->card->reader);
}

//from framework-pkcs15.c, changed sc_esteid_decipher parameters -- mf
//unused, untested
static CK_RV
esteid_prkey_unwrap(struct sc_pkcs11_session *ses, void *obj,
		CK_MECHANISM_PTR pMechanism,
		CK_BYTE_PTR pData, CK_ULONG ulDataLen,
		CK_ATTRIBUTE_PTR pTemplate, CK_ULONG ulAttributeCount,
		void **result)
{
	struct esteid_prkey_object *prkey;
	struct sc_esteid_card *p15card;
	u8	unwrapped_key[256];
	int	rv;

	debug(context, "Initiating key unwrap.\n");

	if (pMechanism->mechanism != CKM_RSA_PKCS)
		return CKR_MECHANISM_INVALID;

	p15card = (struct sc_esteid_card*) ses->slot->card->fw_data;
	prkey = (struct esteid_prkey_object *) obj;
				 
	rv = sc_esteid_decipher(p15card, prkey->prkey_info,
				 pData, ulDataLen,
				 unwrapped_key, sizeof(unwrapped_key));			 
				 
	debug(context, "Key unwrap complete. Result %d.\n", rv);

	if (rv < 0)
		return sc_to_cryptoki_error(rv, ses->slot->card->reader);
	return sc_pkcs11_create_secret_key(ses,
			unwrapped_key, rv,
			pTemplate, ulAttributeCount,
			(struct sc_pkcs11_object **) result);
}

struct sc_pkcs11_object_ops esteid_prkey_ops = {
	NULL,
	NULL,
	esteid_prkey_get_attribute,
	sc_pkcs11_any_cmp_attribute,
	NULL,
	NULL,
	esteid_prkey_sign,
	esteid_prkey_unwrap
};

/*
 * EstEID RSA Public Key Object (as part of certificate)
 */

//from framework-pkcs15.c -- mf
CK_RV esteid_cert_key_get_attribute(struct sc_pkcs11_session *session,
				void *object,
				CK_ATTRIBUTE_PTR attr)
{
	struct esteid_cert_key_object *pubkey = (struct esteid_cert_key_object*) object;
	debug(context,"called\n");
	switch (attr->type) {
	case CKA_CLASS:
		check_attribute_buffer(attr, sizeof(CK_OBJECT_CLASS));
		*(CK_OBJECT_CLASS*)attr->pValue = CKO_PUBLIC_KEY;
		break;
	case CKA_TOKEN:
	case CKA_LOCAL:
	case CKA_SENSITIVE:
	case CKA_ALWAYS_SENSITIVE:
	case CKA_NEVER_EXTRACTABLE:
		check_attribute_buffer(attr, sizeof(CK_BBOOL));
		*(CK_BBOOL*)attr->pValue = TRUE;
		break;
	case CKA_PRIVATE:
	case CKA_MODIFIABLE:
	case CKA_ENCRYPT:
	case CKA_VERIFY:
	case CKA_VERIFY_RECOVER:
	case CKA_WRAP:
	case CKA_DERIVE:
	case CKA_EXTRACTABLE:
		check_attribute_buffer(attr, sizeof(CK_BBOOL));
		*(CK_BBOOL*)attr->pValue = FALSE;
		break;
	case CKA_LABEL:
		check_attribute_buffer(attr, strlen(pubkey->certificate_object->label));
                memcpy(attr->pValue, pubkey->certificate_object->label, strlen(pubkey->certificate_object->label));
		break;
	case CKA_KEY_TYPE:
		check_attribute_buffer(attr, sizeof(CK_KEY_TYPE));
                *(CK_KEY_TYPE*)attr->pValue = CKK_RSA;
		break;
	case CKA_ID:
		check_attribute_buffer(attr, pubkey->certificate_info->id.len);
		memcpy(attr->pValue, pubkey->certificate_info->id.value, pubkey->certificate_info->id.len);
		break;
	case CKA_KEY_GEN_MECHANISM:
		check_attribute_buffer(attr, sizeof(CK_MECHANISM_TYPE));
                *(CK_MECHANISM_TYPE*)attr->pValue = CK_UNAVAILABLE_INFORMATION;
		break;
	case CKA_MODULUS:
		return get_modulus(pubkey->key, attr);
	case CKA_MODULUS_BITS:
		return get_modulus_bits(pubkey->key, attr);
	case CKA_PUBLIC_EXPONENT:
		return get_public_exponent(pubkey->key, attr);
	default:
		return CKR_ATTRIBUTE_TYPE_INVALID;
	}
	return CKR_OK;
}

struct sc_pkcs11_object_ops esteid_cert_key_ops = {
	NULL,
	NULL,
	esteid_cert_key_get_attribute,
	sc_pkcs11_any_cmp_attribute,
	NULL,
	NULL,
	NULL
};

/*
 * EstEID RSA Public Key Object (stored on card as-is)
 * well, there are no pubkeys on the esteid card, 
 * but just in case i left this in -- mf
 */

//from framework-pkcs15.c -- mf
CK_RV esteid_pubkey_get_attribute(struct sc_pkcs11_session *session,
				void *object,
				CK_ATTRIBUTE_PTR attr)
{
	struct esteid_pubkey_object *pubkey = (struct esteid_pubkey_object*) object;

	debug(context,"called\n");
	switch (attr->type) {
	case CKA_CLASS:
		check_attribute_buffer(attr, sizeof(CK_OBJECT_CLASS));
		*(CK_OBJECT_CLASS*)attr->pValue = CKO_PUBLIC_KEY;
                break;
	case CKA_TOKEN:
	case CKA_LOCAL:
	case CKA_SENSITIVE:
	case CKA_ALWAYS_SENSITIVE:
	case CKA_NEVER_EXTRACTABLE:
		check_attribute_buffer(attr, sizeof(CK_BBOOL));
		*(CK_BBOOL*)attr->pValue = TRUE;
                break;
	case CKA_PRIVATE:
	case CKA_MODIFIABLE:
	case CKA_ENCRYPT:
	case CKA_VERIFY:
	case CKA_VERIFY_RECOVER:
	case CKA_WRAP:
	case CKA_DERIVE:
	case CKA_EXTRACTABLE:
		check_attribute_buffer(attr, sizeof(CK_BBOOL));
		*(CK_BBOOL*)attr->pValue = FALSE;
                break;
	case CKA_LABEL:
		check_attribute_buffer(attr, strlen(pubkey->pubkey_object->label));
                memcpy(attr->pValue, pubkey->pubkey_object->label, strlen(pubkey->pubkey_object->label));
		break;
	case CKA_KEY_TYPE:
		check_attribute_buffer(attr, sizeof(CK_KEY_TYPE));
                *(CK_KEY_TYPE*)attr->pValue = CKK_RSA;
                break;
	case CKA_ID:
		check_attribute_buffer(attr, pubkey->pubkey_info->id.len);
		memcpy(attr->pValue, pubkey->pubkey_info->id.value, pubkey->pubkey_info->id.len);
                break;
	case CKA_KEY_GEN_MECHANISM:
		check_attribute_buffer(attr, sizeof(CK_MECHANISM_TYPE));
                *(CK_MECHANISM_TYPE*)attr->pValue = CK_UNAVAILABLE_INFORMATION;
		break;
	case CKA_MODULUS:
		return get_modulus(pubkey->key, attr);
	case CKA_MODULUS_BITS:
		return get_modulus_bits(pubkey->key, attr);
	case CKA_PUBLIC_EXPONENT:
		return get_public_exponent(pubkey->key, attr);
	default:
                return CKR_ATTRIBUTE_TYPE_INVALID;
	}

        return CKR_OK;
}

struct sc_pkcs11_object_ops esteid_pubkey_ops = {
	NULL,
	NULL,
	esteid_pubkey_get_attribute,
	sc_pkcs11_any_cmp_attribute,
	NULL,
	NULL,
	NULL
};

/*
 * get_attribute helpers
 */
 
//from framework-pkcs15.c -- mf
static int 
get_bignum(sc_esteid_bignum_t *bn, CK_ATTRIBUTE_PTR attr)
{
	check_attribute_buffer(attr, bn->len);
	memcpy(attr->pValue, bn->data, bn->len);
	return CKR_OK;
}

//from framework-pkcs15.c -- mf
static int 
get_bignum_bits(sc_esteid_bignum_t *bn, CK_ATTRIBUTE_PTR attr)
{
	CK_ULONG	bits, mask;

	bits = bn->len * 8;
	for (mask = 0x80; mask; mask >>= 1, bits--) {
		if (bn->data[0] & mask)
			break;
	}
	check_attribute_buffer(attr, sizeof(bits));
	*(CK_ULONG *) attr->pValue = bits;
	return CKR_OK;
}

//from framework-pkcs15.c -- mf
static int
get_modulus(struct sc_esteid_pubkey *key, CK_ATTRIBUTE_PTR attr)
{
	if (key == NULL)
		return CKR_ATTRIBUTE_TYPE_INVALID;
	switch (key->algorithm) {
	case SC_ALGORITHM_RSA:
		return get_bignum(&key->u.rsa.modulus, attr);
	}
	return CKR_ATTRIBUTE_TYPE_INVALID;
}

//from framework-pkcs15.c -- mf
static int
get_modulus_bits(struct sc_esteid_pubkey *key, CK_ATTRIBUTE_PTR attr)
{
	if (key == NULL)
		return CKR_ATTRIBUTE_TYPE_INVALID;
	switch (key->algorithm) {
	case SC_ALGORITHM_RSA:
		return get_bignum_bits(&key->u.rsa.modulus, attr);
	}
	return CKR_ATTRIBUTE_TYPE_INVALID;
}

//from framework-pkcs15.c -- mf
static int
get_public_exponent(struct sc_esteid_pubkey *key, CK_ATTRIBUTE_PTR attr)
{
	if (key == NULL)
		return CKR_ATTRIBUTE_TYPE_INVALID;
	switch (key->algorithm) {
	case SC_ALGORITHM_RSA:
		return get_bignum(&key->u.rsa.exponent, attr);
	}
	return CKR_ATTRIBUTE_TYPE_INVALID;
}

//from framework-pkcs15.c -- mf
static int
asn1_sequence_wrapper(const u8 *data, size_t len, CK_ATTRIBUTE_PTR attr)
{
	u8		*dest;
	unsigned int	n;
	size_t		len2;

	len2 = len;
	check_attribute_buffer(attr, len + 1 + sizeof(len));

	dest = (u8 *) attr->pValue;
	*dest++ = 0x30;	/* SEQUENCE tag */
	if (len <= 127) {
		*dest++ = len;
	} else {
		for (n = 4; (len & 0xFF000000) == 0; n--)
			len <<= 8;
		*dest++ = 0x80 + n;
		while (n--) {
			*dest++ = len >> 24;
			len <<= 8;
		}
	}
	memcpy(dest, data, len2);
	attr->ulValueLen = (dest - (u8 *) attr->pValue) + len2;
	debug(context, "pValue = %s\n",attr->pValue);
	return CKR_OK;
}

//from framework-pkcs15.c -- mf
static void
cache_pin(void *p, int user, const void *pin, size_t len)
{
	struct esteid_slot_data *data = (struct esteid_slot_data *) p;

	if (user != 0 && user != 1)
		return;
	memset(data->pin + user, 0, sizeof(data->pin[user]));
	if (len && len <= MAX_CACHE_PIN) {
		memcpy(data->pin[user].value, pin, len);
		data->pin[user].len = len;
	}
}
